import os
import tempfile
import unittest
from pathlib import Path

from smtpymailer.html_parse import convert_html_to_plain_text
from smtpymailer.utils import (
    is_file_with_path,
    ensure_list,
    recipients_to_str,
    convert_bool,
    find_project_root,
    build_all_recipients_and_validate,
    construct_mime_object,
)

from email_validator import EmailNotValidError
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.text import MIMEText

class TestCreateMime(unittest.TestCase):

    # Should return a MIMEImage object if the attachment is an image
    def test_image(self):
        path = os.path.abspath("./tests/assets/dog.jpg")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEImage))

    # Should return a MIMEApplication object if the attachment is a docx file
    def test_docx(self):
        path = os.path.abspath("./tests/assets/dog.docx")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEApplication))

    # Should return a MIMEApplication object if the attachment is a xlsx file
    def test_xlsx(self):
        path = os.path.abspath("./tests/assets/dog.xlsx")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEApplication))

    # Should return a MIMEApplication object if the attachment is an ods file
    def test_ods(self):
        path = os.path.abspath("./tests/assets/dog.ods")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEApplication))

    # Should return a MIMEApplication object if the attachment is an odt file
    def test_odt(self):
        path = os.path.abspath("./tests/assets/dog.odt")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEApplication))

    # Should return a MIMEApplication object if the attachment is a pdf file
    def test_pdf(self):
        path = os.path.abspath("./tests/assets/dog.pdf")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEApplication))

    # Should return a MIMEAudio object if the attachment is a pdf file
    def test_mp3(self):
        path = os.path.abspath("./tests/assets/dog.mp3")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEAudio))

    # Should return a MIMEText object if the attachment is a txt file
    def test_txt(self):
        path = os.path.abspath("./tests/assets/dog.txt")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEText))

    # Should return a MIMEApplication object if the attachment is a zip file
    def test_zip(self):
        path = os.path.abspath("./tests/assets/dog.zip")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEApplication))

    # Should return a MIMEBase object if the attachment is a mp4 file
    def test_base(self):
        path = os.path.abspath("./tests/assets/dog.mp4")
        with open(path, "rb") as attachment:
            part = construct_mime_object(path, attachment)

        self.assertTrue(isinstance(part, MIMEBase))



class TestHtmlConversion(unittest.TestCase):
    def test_convert_html_to_plain_text(self):
        html_input = "<p>Hello,</p><p>Welcome to Python.</p><br>Enjoy learning!<br>"
        expected_output = "Hello,\n\nWelcome to Python.\n\nEnjoy learning!"
        self.assertEqual(convert_html_to_plain_text(html_input), expected_output)

    def test_convert_empty_html_to_plain_text(self):
        html_input = ""
        expected_output = ""
        self.assertEqual(convert_html_to_plain_text(html_input), expected_output)

    def test_convert_html_without_tags_to_plain_text(self):
        html_input = "Hello, Welcome to Python."
        expected_output = "Hello, Welcome to Python."
        self.assertEqual(convert_html_to_plain_text(html_input), expected_output)


class TestIsFileWithPath(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.file_path = os.path.join(self.temp_dir.name, "temp_file")

        with open(self.file_path, "w") as f:
            f.write("Dummy file")

    def tearDown(self):
        self.temp_dir.cleanup()

    def test_is_file_with_path_existing_file(self):
        self.assertTrue(
            is_file_with_path(self.file_path),
            "Expected True for existing file, got False",
        )

    def test_is_file_with_path_non_existing_file(self):
        non_existing_file = os.path.join(self.temp_dir.name, "non_existing_file")
        self.assertFalse(
            is_file_with_path(non_existing_file),
            "Expected False for non-existing file, got True",
        )


class TestEnsureList(unittest.TestCase):
    def test_ensure_list(self):
        """
        Test that `ensure_list` works as expected
        """
        # Scenario 1: Input is already a list
        lst = [1, 2, 3]
        self.assertEqual(ensure_list(lst), [1, 2, 3])

        # Scenario 2: Input is not a list
        not_lst = "string"
        self.assertEqual(ensure_list(not_lst), ["string"])

        # Scenario 3: Input is a numerical value
        num = 42
        self.assertEqual(ensure_list(num), [42])

        # Scenario 4: Input is None
        none_val = None
        self.assertEqual(ensure_list(none_val), [])

        # Scenario 5: Input is a complex type
        complex_type = {"key": "value"}
        self.assertEqual(ensure_list(complex_type), [{"key": "value"}])


class TestRecipientsToStr(unittest.TestCase):
    def test_empty_recipients(self):
        recipients = []
        expected_output = ""
        actual_output = recipients_to_str(recipients)
        self.assertEqual(actual_output, expected_output)

    def test_single_recipient(self):
        recipients = ["test@example.com"]
        expected_output = "test@example.com"
        actual_output = recipients_to_str(recipients)
        self.assertEqual(actual_output, expected_output)

    def test_multiple_recipients(self):
        recipients = ["test1@example.com", "test2@example.com", "test3@example.com"]
        expected_output = "test1@example.com, test2@example.com, test3@example.com"
        actual_output = recipients_to_str(recipients)
        self.assertEqual(actual_output, expected_output)

    def test_single_recipient_as_string(self):
        recipients = "test@example.com"
        expected_output = "test@example.com"
        actual_output = recipients_to_str(recipients)
        self.assertEqual(actual_output, expected_output)

    def test_ensure_list(self):
        recipients = "test@example.com"
        expected_output = ["test@example.com"]
        self.assertEqual(ensure_list(recipients), expected_output)


class TestConvertBool(unittest.TestCase):
    def test_convert_bool(self):
        # Test values that should return True
        true_values = ["True", "true", 1, "1", "Yes", "yes", "Y", "y", True]
        for val in true_values:
            self.assertTrue(convert_bool(val))

        # Test values that should return False
        negative_values = ["False", "2", "No", "no", "N", "n", False, "", []]
        for val in negative_values:
            self.assertFalse(convert_bool(val))

        # Test value that doesn't match either category
        self.assertFalse(convert_bool("unexpected_value"))


class TestFindProjectRoot(unittest.TestCase):
    def test_find_project_root_default(
        self,
    ):
        result = find_project_root()

        # Assert
        self.assertTrue(isinstance(result, Path))

    def test_find_project_root_custom_str_marker(self):
        # Arrange
        result = find_project_root("setup.py")
        # Assert
        self.assertTrue(isinstance(result, Path))

    def test_find_project_root_custom_list_marker(self):
        # Arrange
        result = find_project_root(["foo.bar", "setup.py"])
        # Assert
        self.assertTrue(isinstance(result, Path))

    def test_find_project_root_custom_invalid_marker(self):
        # Arrange
        result = find_project_root("foo.bar")
        # Assert
        self.assertIsNone(result)


class TestBuildAllRecipientsAndValidate(unittest.TestCase):
    #  Should return a list of all recipients after validation
    def test_with_valid_recipients(self):
        recipients = ["test1@example.com", "test2@example.com"]
        cc_recipients = ["test3@example.com"]
        bcc_recipients = ["test4@example.com"]
        expected_result = [
            "test1@example.com",
            "test2@example.com",
            "test3@example.com",
            "test4@example.com",
        ]

        result = build_all_recipients_and_validate(
            recipients, cc_recipients, bcc_recipients
        )
        self.assertEqual(result, expected_result)

    #  Should return an empty list if no recipients are provided
    def test_with_empty_recipients(self):
        recipients = []
        cc_recipients = []
        bcc_recipients = []
        expected_result = []

        with self.assertRaises(ValueError):
            build_all_recipients_and_validate(recipients, cc_recipients, bcc_recipients)

    #  Should return an  list if all recipients are None
    def test_with_none_recipients(self):
        recipients = None
        cc_recipients = None
        bcc_recipients = None

        with self.assertRaises(ValueError):
            build_all_recipients_and_validate(recipients, cc_recipients, bcc_recipients)

    #  Should raise an EmailNotValidError if any of the recipients are not valid email addresses
    def test_with_invalid_email(self):
        recipients = ["invalidemail"]
        with self.assertRaises(EmailNotValidError):
            build_all_recipients_and_validate(recipients, None, None)

    #  Should raise an EmailNotValidError if any of the CC recipients are not valid email addresses
    def test_with_invalid_cc_email(self):
        recipients = ["test1@example.com"]
        cc_recipients = ["invalidemail"]
        with self.assertRaises(EmailNotValidError):
            build_all_recipients_and_validate(recipients, cc_recipients, None)

    #  Should raise an EmailNotValidError if any of the BCC recipients are not valid email addresses
    def test_with_invalid_bcc_email(self):
        recipients = ["test1@example.com"]
        bcc_recipients = ["invalidemail"]
        with self.assertRaises(EmailNotValidError):
            build_all_recipients_and_validate(recipients, None, bcc_recipients)




if __name__ == "__main__":
    unittest.main()



import os
import pathlib
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate, make_msgid
from typing import Union, Optional, List
import dns.resolver


# Create a custom resolver and specify nameservers directl
custom_resolver = dns.resolver.Resolver()
custom_resolver.nameservers = ['8.8.8.8', '8.8.4.4']

# Override the default resolver
dns.resolver.default_resolver = custom_resolver

import pydig
from dotenv import load_dotenv

from smtpymailer.html_parse import (
    convert_html_to_plain_text,
    convert_img_elements,
    make_html_content
)
from smtpymailer.utils import (
    is_file_with_path,
    convert_bool,
    recipients_to_str,
    build_all_recipients_and_validate,
    ensure_list,
    construct_mime_object,
)
from smtpymailer.validation import (
    validate_user_email,
    validate_dmarc_record,
    spf_check,
    get_address_type,
    validate_dkim_record,
)


# Create a custom resolver and specify nameservers directl
custom_resolver = dns.resolver.Resolver()
custom_resolver.nameservers = ['8.8.8.8', '8.8.4.4']

# Override the default resolver
dns.resolver.default_resolver = custom_resolver

class Contact:
    """
    Contact class represents a contact with an email address and a name. It provides methods to manipulate and retrieve information about the contact.

    Attributes:
        email (str): The contact's email address.
        name (Optional[str]): The contact's name (optional).

    Methods:
        __init__(self, email: str, name: Optional[str] = None, validate_email: bool = True):
            Initializes a Contact object with the specified email address and name.
            Raises EmailNotValidError if the email address is not valid.

        __repr__(self) -> str:
            Returns a string representation of the Contact object.

        get_domain(self) -> str:
            Returns the domain of the email address.

    Raises:
        EmailNotValidError: If the email address is not valid.
    """

    email: str
    name: Optional[str]

    def __init__(self, email, name, validate_email: bool = True):
        """
        Creates a contact object to send an email from or to
        Args:
            email (str): email address
            name optional(str): name
            validate_email (bool, optional): whether to validate the email address

        Raises:
            EmailNotValidError: if the email address is not valid

        """
        self.email = email
        self.name = name

        if validate_email:
            validate_user_email(self.email)

    def __repr__(self):
        if self.name:
            return f"{self.name} <{self.email}>"
        else:
            return f"{self.email}"

    def get_domain(self):
        """
        Gets the domain of the email address
        Returns:
            str: domain

        """
        return self.email.split("@")[1]


def validate_send_email(
    html_content,
    template,
    template_directory,
    subject,
    recipients,
):
    """
    Validates the parameters for the `send_email` function. it performs the following validations:
        * If neither `html_content` nor `template` is provided.
        * If `template` is a file name but `template_directory` is not provided.
        * If `subject` is empty.
        * If `recipients` is not provided.

    Args:
        html_content (Optional(str)): The HTML content of the email.
        template (Optional(str)): The name of the template to use for the email. If `html_content` is not provided.
            This should be a valid Jinja template.
        template_directory (Optional(Union[list,str])): The directory where the template file is located (if using a template).
            This should only be provided if `template` is a file name and not a full path. Can be a string or list of
            strings.
        subject (str): The subject of the email.
        recipients (str or list): The recipient(s) of the email. This can be a single email address as a string,
            or a list of email addresses.

    Raises:
        ValueError: If neither `html_content` nor `template` is provided.
        FileNotFoundError: If `template` is a file name but `template_directory` is not provided.
        ValueError: If `subject` is empty.
        ValueError: If `recipients` is not provided.

    """
    if not html_content or not template:
        if template:
            if not is_file_with_path(template) and not template_directory:
                raise FileNotFoundError(
                    "Template file not found. Please provide either a full path to the template or a template file and template directory"
                )
            elif not is_file_with_path(template) and not any(
                [
                    is_file_with_path(os.path.join(x, template))
                    for x in ensure_list(template_directory)
                ]
            ):
                raise FileNotFoundError(
                    "Template file not found. Please provide either a full path to the template or a template file and template directory"
                )
        if not html_content and not template:
            raise ValueError(
                "Please provide either 'html_content' or a jinja template to render"
            )

    if not subject:
        raise ValueError("Subject cannot be empty. Please provide a subject")

    if not recipients:
        raise ValueError(
            "Recipients not provided. Please provide a single recipient's email or a list of email addresses"
        )


class SmtpMailer:
    """
    Send emails from alternative domains names to the mail server. DNS records must be correctly assigned to the
    sending domain.
    """

    sender: Contact
    html_content: Optional[Union[str, pathlib.Path]] = None
    mail_server: str
    mail_port: int
    mail_use_tls: bool
    mail_username: str
    mail_password: str
    mail_dkim_selector: str = "mail"
    message: MIMEMultipart
    message_alt: MIMEMultipart

    def __init__(self, sender_email: str, sender_name: Optional[str], **kwargs):
        """
        Send emails from alternative domains names to the mail server. DNS records must be correctly assigned to the
        sending domain.

        Required DNS must include:

        * A valid SPF record
        * A valid DMARC record
        * A DKIM record generated from YOUR sending mail server

        Email server auth details need to be supplied, and in the interest of security can and are preferred to be
        stored in the environment. `SmtpMailer` will check the `os.environ`, dotenv `.env` file in the root directory
        or can be supplied (**IF YOU HAVE TO**) as kwargs.

        Notes:
            If either the email fails to validate or the email server auth fails to validate, an exception will be
            raised.

        Configuration Parameters:
            **MAIL_SERVER**: The hostname or IP address of the email server.

            **MAIL_PORT**: The port number of the email server.

            **MAIL_USE_TLS**: Boolean to enable/disable Transport Layer Security.

            **MAIL_USERNAME**: The username for the email server.

            **MAIL_PASSWORD**: The password for the email server.

            **MAIL_DKIM_SELECTOR**: The DKIM selector to use for email signing, created by the mail server and applied
            to the sending domain's DNS records.


        Args:
            sender_email (str): Email address to send from
            sender_name (str): Name to use in email from field

        """

        self.sender = Contact(sender_email, sender_name)
        self._setup_email_server_auth(**kwargs)
        self._validate_auth_setup()

    def _validate_auth_setup(self):
        """
        Validates the setup of the email server. Checks for the following:
            - sender - validated email
            - SPF record for your mail server
            - DKIM record for your mail server
            - DMARC record for your mail server

        Returns:
            bool

        Raises:
            Exception: if the setup is not valid

        """
        resolver = pydig.Resolver(
            nameservers=["1.1.1.1", "1.0.0.1", "8.8.8.8", "8.8.4.4"]
        )
        sender_domain = self.sender.get_domain()

        dmarc_records, spf_records, dkim_records = self._query_dns_records(
            sender_domain
        )

        self._validate_records(sender_domain, dmarc_records, spf_records, dkim_records)

    def _query_dns_records(self, sender_domain):
        """
        Args:
            sender_domain: The domain name of the sender for which DNS records need to be queried.

        Returns:
            A tuple of three lists representing the DNS records:
            - dmarc_records: A list of TXT records associated with the "_dmarc" subdomain of the sender domain.
            - spf_records: A list of TXT records that contain "spf" keyword and are associated with the sender domain.
            - dkim_records: A list containing a single string, which is the concatenated DKIM record.
        """

        dmarc_records = []
        spf_records = []
        dkim_record = ""

        dmarc_answers = dns.resolver.resolve(f"_dmarc.{sender_domain}", "TXT")
        for rdata in dmarc_answers:
            dmarc_records.append(rdata.strings[0].decode())

        spf_answers = dns.resolver.resolve(sender_domain, "TXT")
        for rdata in spf_answers:
            if any("spf" in txt.decode() for txt in rdata.strings):
                spf_records.append(rdata.strings[0].decode())

        dkim_answers = dns.resolver.resolve(f"{self.mail_dkim_selector}._domainkey.{sender_domain}", "TXT")
        for rdata in dkim_answers:
            dkim_record = ''.join(txt.decode() for txt in rdata.strings)

        return dmarc_records, spf_records, [dkim_record] if dkim_record else []

    def _validate_records(
        self,
        sender_domain: str,
        dmarc_records: list,
        spf_records: list,
        dkim_records: list,
    ):
        """
        Args:
            sender_domain (str): The domain of the sender.
            dmarc_records (list): A list of DMARC records for the sender domain.
            spf_records (list): A list of SPF records for the sender domain.
            dkim_records (list): A list of DKIM records for the sender domain.

        Raises:
            Exception: If none of the DMARC records for the sender domain are valid.
            Exception: If none of the SPF records for the sender domain are valid.
            Exception: If none of the DKIM records for the sender domain are valid.
        """

        valid_dmarc = False
        for record in dmarc_records:
            if validate_dmarc_record(record):
                valid_dmarc = True
                break
        if not valid_dmarc:
            raise Exception(f"No valid DMARC record found for {sender_domain}")

        valid_spf = False
        for record in spf_records:
            if spf_check(record, **get_address_type(self.mail_server)):
                valid_spf = True
                break
        if not valid_spf:
            raise Exception(f"No valid SPF record found for {sender_domain}")

        valid_dkim = False
        for record in dkim_records:
            if validate_dkim_record(record):
                valid_dkim = True
                break
        if not valid_dkim:
            raise Exception(f"No valid DKIM record found for {sender_domain}")

    def _setup_email_server_auth(self, **kwargs):
        """
        Args:
            **kwargs: Additional keyword arguments that can be passed to the method.

        Raises:
            Exception: If a required configuration value is not found.

        """
        # Load .env file if it exists
        load_dotenv()

        # Helper function to get the configuration value
        def get_config(key):
            """
            Args:
                key: The config key to retrieve the value for.

            Returns:
                The value corresponding to the given config key, or None if the key was not found in any of the
                available sources (dotenv, environment variables, or kwargs).

            """

            # This is discouraged and only first, so I force a unit test to fail.
            key_lower = key.lower()
            for k, v in kwargs.items():
                if k.lower() == key_lower:
                    return v

            # Check in dotenv
            value = os.getenv(key) or os.getenv(key.lower())
            if value:
                return value

            raise ValueError(f"Could not find config value for {key}, cannot continue.")

        # Configuration values
        self.mail_server = get_config("MAIL_SERVER")
        self.mail_port = int(get_config("MAIL_PORT"))
        self.mail_use_tls = convert_bool(get_config("MAIL_USE_TLS"))
        self.mail_username = get_config("MAIL_USERNAME")
        self.mail_password = get_config("MAIL_PASSWORD")
        self.mail_dkim_selector = get_config("MAIL_DKIM_SELECTOR")

        # check for required values and raise exception if not found
        with self._connect_to_server() as server:
            server.quit()

    def _connect_to_server(self):
        """
        Connects to the mail server.
        Returns:
            server: An instance of the server connection
        Raises:
            ValueError: If there is an error connecting to the mail server or invalid server connection details
        """
        try:
            server = smtplib.SMTP(self.mail_server, self.mail_port)
            if self.mail_use_tls:
                server.starttls()
            server.login(self.mail_username, self.mail_password)
            return server
        except:
            raise ValueError(
                "Invalid server connection details, please check and try again"
            )

    def _send_message(self, recipients: list):
        """
        Sends a message if provided.

        Args:
            recipients: a list of all email addresses to send email to including bcc emails that are not in the header.

        Returns:
            True if the message was sent successfully, otherwise raises an Exception.
        """
        try:
            with self._connect_to_server() as server:
                server.sendmail(str(self.sender), recipients, self.message.as_string())
            return True
        except Exception as e:
            raise Exception(f"Failed to send message: {e}")

    def _construct_base_message(
        self,
        subject: str,
        recipients: str,
        cc_recipients: Optional[str] = None,
        reply_to: Optional[str] = None,
    ):
        """
        Args:
            subject (str): The subject of the email message.
            recipients (str): The recipients of the email message.
            cc_recipients (Optional[str], optional): The CC recipients of the email message. Defaults to None.
            reply_to (Optional[str], optional): The reply-to address of the email message. Defaults to None.
            None.

        Returns:
            MIMEMultipart: The constructed base email message.

        """
        message = MIMEMultipart("mixed")

        message["Subject"] = subject
        message["To"] = recipients
        message["Date"] = formatdate(localtime=True)
        message["Message-Id"] = make_msgid()
        message["From"] = str(self.sender)

        if cc_recipients:
            message["Cc"] = cc_recipients

        if reply_to and validate_user_email(email=reply_to, check_deliverability=True):
            message["Reply-To"] = reply_to

        return message

    def _add_attachments(self, attachments: Optional[Union[List[str], str]] = None):
        """
        Args:
            attachments: Optional[List[str] or str] - The attachment file paths to be added. It can be a single file path string or a list of file paths.

        Raises:
            Exception:
                * If the attachment file path is invalid or non-existent.
                * If the file is not found.
                * If there is a permission error while opening the file.
                * If there is an error while opening the attachment file.
        Returns:
            None

        """
        attachments = ensure_list(attachments)

        for attachment_file_path in attachments:

            # Expand the user path, if needed and user has used ~/path/to/file
            attachment_file_path = os.path.expanduser(attachment_file_path)

            if isinstance(attachment_file_path, str) and os.path.isfile(
                attachment_file_path
            ):
                try:
                    with open(attachment_file_path, "rb") as attachment:
                        part = construct_mime_object(attachment_file_path, attachment)
                        part.add_header(
                            "Content-Disposition",
                            f"attachment; filename={os.path.basename(attachment_file_path)}",
                        )
                        self.message.attach(part)

                except PermissionError:
                    raise Exception(
                        f"Cannot open file. Check if the file is open in another program {attachment_file_path}"
                    )
                except IOError as e:
                    raise Exception(
                        f"Error opening attachment file {attachment_file_path}: {e}"
                    )
            else:
                raise FileNotFoundError(
                    f"Invalid or non existent attachment file path: {attachment_file_path}"
                )

    def _make_plain_message(self, html_content):
        """
        Creates a plain text version of the email, using the HTML content it coverts it to plain text.

        Args:
            html_content: A string representing the HTML content of the email.

        Notes:
            This method sets the class attribute `message_alt` to a MIMEMultipart object containing the plain
            text version of the email.

        """
        plain_html = convert_html_to_plain_text(html_content)

        # Attach the plain and HTML versions
        self.message_alt = MIMEMultipart("alternative")
        self.message.attach(self.message_alt)

        # create plain email
        msg_text = MIMEText(plain_html, "plain")
        self.message_alt.attach(msg_text)

    def _make_html_message(self, html_content: str):
        """
        Creates an HTML version of the email, using the HTML content it converts it to HTML.

        It may also alter the HTML content based on data attributes specified in the img tags.

        Args:
            html_content (str): The HTML content of the message.

        Notes:
            This method sets the class attribute `message_alt` to a MIMEMultipart object containing the HTML
            version of the email.

        """

        html_content = convert_img_elements(html_content, self.message)

        # create html email
        msg_html = MIMEText(html_content, "html")
        self.message_alt.attach(msg_html)

    def send_email(
        self,
        recipients: Union[str, list],
        subject: str,
        cc_recipients: Optional[Union[str, list]] = None,
        bcc_recipients: Optional[Union[str, list]] = None,
        reply_to: Optional[str] = None,
        attachments: Optional[Union[List, str]] = None,
        html_content: Optional = None,
        template: Optional = None,
        template_directory: Optional[Union[str, list]] = None,
        **kwargs,
    ):
        """
        Sends an email with the given parameters. The email can be sent with either HTML content or a Jinja template.

        You can convert image src's (files or URLs) to base64 embedded images or attach them as CID images in the email.
        This can easily be achieved by adding a data attribute to the image.

            - `data-cid` will convert the image to a CID image and attach to the email inline.
            - `data-base` will convert the image to a base64 embedded image.

        Args:
            recipients: Either a single recipient email address or a list of email addresses.
            subject: The subject of the email.
            cc_recipients: Optional. Either a single cc recipient email address or a list of email addresses.
            bcc_recipients: Optional. Either a single bcc recipient email address or a list of email addresses.
            reply_to: Optional. The reply-to email address.
            attachments: Optional. Either a single attachment file path or a list of attachment file paths.
            html_content: Optional. The HTML content of the email, not needed if you are using a template.
            template: Optional. The template file path.
            template_directory: Optional. Either a single template directory or a list of template directories.
            **kwargs: Additional keyword arguments for the jinja template if needed

        Raises:
            Exception:
                * If neither 'html_content' nor a template is provided
                * When a `template` str is passed with no directory and no `template_directory` is supplied.
                * Blank subject field
                * Empty str or list of `recipients`

        Returns:
            True if the email was sent successfully, otherwise raises an Exception.


        """

        validate_send_email(
            html_content, template, template_directory, subject, recipients
        )

        all_recipients = build_all_recipients_and_validate(
            recipients, cc_recipients, bcc_recipients
        )

        recipients_str = recipients_to_str(recipients)
        cc_recipients_str = recipients_to_str(cc_recipients)

        self.message = self._construct_base_message(
            recipients=recipients_str,
            subject=subject,
            cc_recipients=cc_recipients_str,
            reply_to=reply_to,
        )

        html_content = make_html_content(
            html_content,
            template,
            template_directory,
            **kwargs,
        )

        self._make_plain_message(html_content)
        self._make_html_message(html_content)
        self._add_attachments(attachments)

        return self._send_message(all_recipients)



import base64
import datetime
import hashlib
import os
import re
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from io import BytesIO
from typing import Optional, List, Union
from urllib.parse import urlparse
import requests
from bs4 import BeautifulSoup
from bs4.element import Tag
from jinja2 import Environment, FileSystemLoader, select_autoescape
import html2text
from PIL import Image

from smtpymailer.utils import is_get_local_file, is_resolvable, guess_type_by_extension


def check_data_in_html_el(html_content: str):
    """
    Check the presence of specific data attributes in HTML content

    Args:
        html_content (str): The HTML content to check.

    Returns:
        dict: A dictionary containing the count of each data attribute found.
            - "base64": The count of base64 encoded images ("data:image/jpeg;base64,").
            - "cid": The count of images with the attribute "cid:smtpymailer-image" in the source.
            - "data-smtpymailer": The count of images with the data attribute data-smtpymailer in the source.

    Example:
        >>> html_content = '<img src="data:image/jpeg;base64,abc123" /> <img src="cid:smtpymailer-image1.jpg" />'
        >>> check_data_in_html_el(html_content)
        {'base64': 1, 'cid': 1, 'data-smtpymailer': 0}
    """
    soup = BeautifulSoup(html_content, "html.parser")

    data_found = {"base64": 0, "cid": 0, "data-smtpymailer": 0}

    for img in soup.find_all("img"):
        src = img.get("src", "")
        data_found["base64"] += 1 if ";base64," in src else 0
        data_found["cid"] += 1 if "cid:smtpymailer-image" in src else 0
        data_found["data-smtpymailer"] += 1 if "data-smtpymailer" in img.attrs else 0

    return data_found


def change_image_type(element: "Tag", mime_type: str, img_data: bytes):
    """
    Optionally changes the image type of the given image data, based on data-attributes in the HTML img element.
    See Notes for more information.

    Args:
        element (Tag): The HTML img element.
        mime_type (str): The MIME type of the image.
        img_data (bytes): The image data.

    Returns:
        tuple: A tuple containing the converted image data and the converted MIME type.

    Notes:
        I found a wierd bug in gmail showing some CID inline png images as `noname` attachments, I had 4 CID
        attachments and only one was showing as `noname`. This was in the gmail preview only, not visible when the
        email was viewed.

        To fix, I converted that image into a jpg and the 'noname' attachment disappeared. There is very little online
        about this, but possible a bug in gmail.

        To use this feature, add the following data attributes to the img element:

            * data-convert="jpg" - Converts the image type if it's not already. Can be "png", "jpg", or "gif".
            * data-format="rgb" (optional): The pixel format of the image. Defaults to "RGB". Can also be "RGBA".

        i.e  <img data-convert="jpg" data-format="rgb" src="https://www.example.com/image.png"/>

    """

    if "data-convert" in element.attrs:
        convert_to_type = element.get("data-convert").lower()
        if convert_to_type not in ["png", "jpg", "gif", "jpeg"]:
            raise ValueError(
                f"Invalid value for data-convert attribute: {convert_to_type}"
            )
        pixel_format = "RGB"
        if "data-format" in element.attrs:
            pixel_format = element.get("data-format").lower()
            if pixel_format not in ["rgb", "rgba"]:
                raise ValueError(
                    f"Invalid value for data-format attribute: {pixel_format}"
                )

        _, file_format = mime_type.split("/")
        file_format = file_format.replace("jpeg", "jpg").lower()

        if convert_to_type != file_format:
            img = Image.open(BytesIO(img_data))
            with BytesIO() as img_io:
                img.convert(pixel_format.upper()).save(img_io, convert_to_type.upper())
                img_data = img_io.getvalue()  # Update img_data with JPEG data
                mime_type = f'image/{convert_to_type.lower().replace("jpg", "jpeg")}'  # Update content_type to JPEG

    return img_data, mime_type


def process_img_element(
        img: Tag,
        idx: int,
        convert_to_base64: bool = False,
        email_message: MIMEMultipart = None,
):
    """
    Process and manipulate HTML img elements, converting the image to either CID attachments or base64 encoded.
    If the request fails the element is ignored and left with the original src.
    Checks for the src in the local filesystem and if its a valid url so both local and remote images can be used.

    Optionally you can convert the original image type when using CID attachments. See Notes `change_image_type` for
    more information.

    Args:
        img (Tag): The HTML img element to process.
        idx (int): The index of the img element.
        convert_to_base64 (bool): Boolean indicating whether to convert the image to base64. Defaults to False.
        email_message (MIMEMultipart): An instance of the EmailMessage class. If provided, the image will be attached
            to the email message as an inline image.

    Notes:
        all below are valid examples:

            > <img src="https://www.example.com/image.png" data-convert="jpg" data-format="rgb" />

            > <img src="/home/user/img.jpg" data-convert='png'/>

    """
    img_data = None

    src = img.get("src", "")
    content_type = guess_type_by_extension(src)

    is_local_file, path = is_get_local_file(src)
    if is_local_file:
        with open(path, "rb") as f:
            img_data = f.read()

    elif is_resolvable(src):
        try:
            response = requests.get(src, stream=True)
            response.raise_for_status()
            img_data = response.content
        except requests.RequestException:
            # if the request fails, ignore the element and leave the original src
            return

    if img_data:

        img_data, content_type = change_image_type(img, content_type, img_data)

        if convert_to_base64:
            parsed_url = urlparse(src)
            img_ext = os.path.splitext(parsed_url.path)[1].lstrip(".")
            base64_data = base64.b64encode(img_data).decode("utf-8")
            img["src"] = f"data:image/{img_ext};base64,{base64_data}"
            img["data-smtpymailer"] = ""

        elif email_message is not None:
            hash_object = hashlib.md5(img_data)
            cid = hash_object.hexdigest() + f"{idx}"
            maintype, subtype = content_type.split("/")

            # Create an instance of MIMEImage
            mime_image = MIMEImage(img_data, _subtype=subtype)
            mime_image.add_header("Content-ID", f"<{cid}>")
            mime_image.add_header("Content-Disposition", "inline")

            # Attach it to the email message
            email_message.attach(mime_image)
            img["src"] = f"cid:{cid}"
            img["data-smtpymailer"] = ""


def convert_img_elements(html_content: str, email: MIMEMultipart) -> str:
    """
    Converts all img elements in the given HTML content to base64.

    Args:
        html_content (str): The HTML content containing img elements.
        email (MIMEMultipart): An instance of the EmailMessage class.

    Returns:
        str: The modified HTML content with img elements converted to base64.

    """
    soup = BeautifulSoup(html_content, "html.parser")
    for idx, img in enumerate(soup.find_all("img")):
        if "data-base" in img.attrs:
            process_img_element(img, idx, convert_to_base64=True)
        elif "data-cid" in img.attrs:
            process_img_element(img, idx, email_message=email)
    return str(soup)


def convert_html_to_plain_text(html_text: str) -> str:
    """
    Converts HTML text to plain text, ignoring images, links, and markdown-style headers,
    and removing excess whitespace.

    Args:
      html_text (str): HTML text

    Returns:
        str: plain text version of the email
    """

    # Create a html2text object
    h = html2text.HTML2Text()

    # Configure html2text
    h.ignore_images = True
    h.ignore_links = True
    h.strong_mark = ""
    h.ul_item_mark = ""
    h.emphasis_mark = ""

    # Convert HTML to plain text
    plain_text = h.handle(html_text)

    # Remove markdown header symbols
    plain_text = re.sub(r"#+ ", "", plain_text)

    # Remove extra whitespace/newlines after paragraph breaks
    plain_text = re.sub(r"\n\s*\n", "\n\n", plain_text)

    return plain_text.strip()


def create_jinja_environment(template_paths: List[str]) -> Environment:
    """
    Creates a Jinja2 environment.

    Args:
        template_paths (List[str]): List of paths where the templates can be found.

    Returns:
        Environment: The Jinja2 Environment object.
    """
    return Environment(
        loader=FileSystemLoader(template_paths),
        autoescape=select_autoescape(["html", "xml"]),
    )


def render_html_template(
        template: str,
        template_paths: Optional[List[str]] = None,
        extension: str = "html",
        **kwargs,
) -> str:
    """
    Renders an HTML template using Jinja2.

    This function takes the name of an HTML template and optional paths where the template can be found.
    It then renders the template into HTML content. Additional context for rendering can be provided
    through keyword arguments.

    Args:
        template (str): The name of the template file to render.
        template_paths (Optional[List[str]], optional): A list of paths where the template files are located.
            If None, a default path is used. Defaults to None.
        extension (str, optional): The file extension of the template file. Defaults to "html".
        **kwargs: Arbitrary keyword arguments that provide context for the template rendering.
            These arguments are passed directly to the Jinja2 template.

    Returns:
        str: The rendered HTML content as a string.

    Examples:
        Render a simple template without additional paths or context:
        >>> render_html_template('my_template')

        Render a template with additional paths and context:
        >>> render_html_template('my_template', template_paths=['/path/to/templates'], username='Alice')
    """

    # If template_paths is not provided, use the path of the template

    if not template_paths:
        split_path = os.path.split(os.path.abspath(template))
        template_paths = [split_path[0]]
        template = split_path[1]
    # Create the Jinja2 environment with the specified paths
    env = create_jinja_environment(template_paths)
    # Get the template and render it with the provided keyword arguments
    template = env.get_template(template)
    html_content = template.render(dated=datetime.datetime.now(), **kwargs)
    return html_content



def make_html_content(
        html_content: Optional[str] = None,
        template: Optional[str] = None,
        template_directory: Optional[Union[str, List[str]]] = None,
        **kwargs,
):
    """
    Adds HTML content to the email message. This function either directly uses provided HTML content or renders HTML content
    from a specified template.

    Args:
        html_content (Optional[str]): A string containing the HTML content of the email. If this is provided, the function
            will use it directly. Defaults to None.
        template (Optional[str]): A string representing the name of the template file to be used for rendering the HTML
            content. This is used if 'html_content' is not provided. Defaults to None.
        template_directory (Optional[Union[str, List[str]]]): A string or list of strings representing the directories
            to search for the template file. This is used in conjunction with 'template' to render the HTML content.
            Defaults to None.
        **kwargs: Additional keyword arguments that are passed to the template rendering function.

    Returns:
        str: The HTML content that will be used in the email. This could be the provided 'html_content' or content rendered
             from the specified template.

    """
    if not html_content:
        html_content = render_html_template(template, template_directory, **kwargs)

    return html_content



import socket
import unittest
from smtpymailer.validation import (
    validate_user_email,
    validate_dmarc_record,
    is_ip_in_network,
    resolve_domain,
    get_address_type,
    validate_dkim_record,
    spf_check,
)


class TestValidation(unittest.TestCase):
    def test_validate_user_email_return_normalized_true(self):
        result = validate_user_email(
            "test@EXAMPLE.com", return_normalized=True, check_deliverability=False
        )
        self.assertEqual(result, "test@example.com")

    def test_validate_user_email_return_normalized_false(self):
        result = validate_user_email(
            "test@EXAMPLE.com", return_normalized=False, check_deliverability=False
        )
        self.assertEqual(result, "test@EXAMPLE.com")

    def test_validate_user_email_invalid_email(self):
        with self.assertRaises(Exception):  # Depending on the expected exception
            validate_user_email(
                "invalid_email", return_normalized=False, check_deliverability=False
            )

    def test_validate_dmarc_record_valid(self):
        valid_dmarc_record = (
            "v=DMARC1; p=none; rua=mailto:abc@example.com; pct=100; fo=1;"
        )
        self.assertIsNotNone(validate_dmarc_record(valid_dmarc_record))

    def test_validate_dmarc_record_invalid(self):
        invalid_dmarc_record = "This is invalid dmarc record"
        with self.assertRaises(ValueError):
            validate_dmarc_record(invalid_dmarc_record)

    def test_validate_dmarc_record_invalid_two(self):
        invalid_dmarc_record = (
            "v=SPF; p=none; rua=mailto:abc@example.com; pct=100; fo=1;"
        )
        with self.assertRaises(ValueError):
            validate_dmarc_record(invalid_dmarc_record)

    def test_validate_dmarc_record_pct_value_out_of_range(self):
        invalid_dmarc_record = (
            "v=DMARC1; p=none; rua=mailto:abc@example.com; pct=200; fo=1;"
        )
        with self.assertRaises(ValueError):
            validate_dmarc_record(invalid_dmarc_record)

    def test_validate_dmarc_regex_error(self):
        invalid_dmarc_record = "v=DMARC1; p=reject; rua=mailto:this_is_not_an_email; ruf=mailto:another_invalid_email@; pct=100"

        with self.assertRaises(ValueError):
            validate_dmarc_record(invalid_dmarc_record)

    def test_validate_dmarc_invalid_email(self):
        invalid_dmarc_record = "v=DMARC1; p=none; rua=mailto:abc; pct=200; fo=1;"
        with self.assertRaises(ValueError):
            validate_dmarc_record(invalid_dmarc_record)

    def test_valid_ipv4_in_network(self):
        self.assertTrue(is_ip_in_network("192.168.1.1", "192.168.1.0/24"))

    def test_valid_ipv4_outside_network(self):
        self.assertFalse(is_ip_in_network("192.168.2.1", "192.168.1.0/24"))

    def test_valid_ipv6_in_network(self):
        self.assertTrue(
            is_ip_in_network(
                "2001:0db8:85a3:0000:0000:8a2e:0370:7334", "2001:0db8::/32"
            )
        )

    def test_invalid_ip(self):
        with self.assertRaises(ValueError):
            is_ip_in_network("not-a-valid-ip", "192.168.1.0/24")

    def test_invalid_network(self):
        with self.assertRaises(ValueError):
            is_ip_in_network("192.168.1.1", "not-a-valid-network")

    def test_empty_string_ip(self):
        with self.assertRaises(ValueError):
            is_ip_in_network("", "192.168.1.0/24")

    def test_empty_string_network(self):
        with self.assertRaises(ValueError):
            is_ip_in_network("192.168.1.1", "")

    def test_ipv4(self):
        # This test checks if IPv4 address is returning as expected
        ipv4, ipv6 = resolve_domain("www.google.com")
        self.assertTrue(all(isinstance(ip, str) for ip in ipv4))
        self.assertTrue(all(ip.count(".") == 3 for ip in ipv4))

    def test_ipv6(self):
        # This test checks if IPv6 address is returning as expected
        ipv4, ipv6 = resolve_domain("www.google.com")
        self.assertTrue(all(isinstance(ip, str) for ip in ipv6))
        self.assertTrue(all(ip.count(":") > 0 for ip in ipv6))

    def test_invalid_domain(self):
        # This test checks if the function raises the appropriate exception for invalid domains
        with self.assertRaises(socket.gaierror):
            ipv4, ipv6 = resolve_domain("invalid.domain")

    def test_ipv4_address(self):
        address = "192.0.2.0"
        expected_result = {"ipv4": address}
        self.assertEqual(get_address_type(address), expected_result)

    def test_ipv6_address(self):
        address = "2001:db8::"
        expected_result = {"ipv6": address}
        self.assertEqual(get_address_type(address), expected_result)

    def test_domain_name(self):
        address = "example.com"
        expected_result = {"domain": address}
        self.assertEqual(get_address_type(address), expected_result)

    def test_invalid_input(self):
        address = "invalid"
        expected_result = {"invalid": address}
        self.assertEqual(get_address_type(address), expected_result)

    def test_validate_dkim_record_valid(self):
        valid_dkim_record = "v=DKIM1; h=sha256; k=rsa; p=MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAzgLBsMFlJoX5XRcgT7T/"
        self.assertEqual(
            validate_dkim_record(valid_dkim_record), True
        )  # change expected output

    def test_validate_dkim_record_invalid(self):
        invalid_dkim_record = "invalid_dkim_record"
        self.assertEqual(validate_dkim_record(invalid_dkim_record), False)

    def test_validate_dkim_record_empty(self):
        empty_dkim_record = ""
        self.assertEqual(validate_dkim_record(empty_dkim_record), False)

    #  Test valid IPv6 address in network
    def test_valid_ipv6_in_network_attempt(self):
        ip = "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
        network = "2001:0db8::/32"

        result = is_ip_in_network(ip, network)
        self.assertTrue(result)

    #  Test valid IPv6 address outside network
    def test_valid_ipv6_outside_network(self):
        ip = "2001:0db8:abcd:ef01:0000:0000:0000:0001"
        network = "2001:0db8:abcd:ef02::/64"

        result = is_ip_in_network(ip, network)
        self.assertFalse(result)

    #  Test invalid type raises TypeError
    def test_invalid_type(self):
        with self.assertRaises(TypeError):
            is_ip_in_network(1, 1)


class TestSpfCheck(unittest.TestCase):
    #  Check if an IP address is authorized by the SPF record.
    def test_ip_address_authorized(self):
        spf_record = "v=spf1 ip4:192.0.2.0/24 -all"
        ip_address = "192.0.2.1"
        result = spf_check(spf_record, ipv4=ip_address)
        self.assertTrue(result)

    #  Check if a domain is authorized by the SPF record.
    def test_domain_authorized(self):
        spf_record = "v=spf1 include:example.com -all"
        domain = "example.com"
        result = spf_check(spf_record, domain=domain)
        self.assertTrue(result)

    #  Check if an IPv4 address is authorized by the SPF record.
    def test_ipv4_address_authorized(self):
        spf_record = "v=spf1 ip4:192.0.2.0/24 -all"
        ipv4_address = "192.0.2.1"
        result = spf_check(spf_record, ipv4=ipv4_address)
        self.assertTrue(result)

    #  Test domain fail
    def test_domain_fail(self):
        spf_record = "v=spf1 include:example.com -all"
        result = spf_check(spf_record=spf_record, domain="yourdomain.com")
        self.assertFalse(result)

    #  Test domain pass
    def test_domain_pass(self):
        spf_record = "v=spf1 include:example.com -all"
        result = spf_check(spf_record=spf_record, domain="example.com")
        self.assertTrue(result)

    #  Check if an invalid SPF record raises a ValueError.
    def test_invalid_spf_record(self):
        spf_record = "invalid_spf_record"
        with self.assertRaises(ValueError):
            spf_check(spf_record)

    #  Check if an invalid IP address returns False.
    def test_invalid_ip_address(self):
        spf_record = "v=spf1 ip4:192.0.2.0/24 -all"
        ipv4_address = "invalid_ip_address"
        with self.assertRaises(ValueError):
            spf_check(spf_record, ipv4=ipv4_address)

    #  Check if an invalid network returns False.
    def test_invalid_network(self):
        spf_record = "v=spf1 ip4:192.0.2.0/24 -all"
        ipv4_address = "195.0.2.1"
        result = spf_check(spf_record, ipv4=ipv4_address)
        self.assertFalse(result)

    def test_spf_check_with_ipv6(self):
        spf_record = "v=spf1 ip6:2001:0db8::/32 -all"
        ipv6_address = "2001:0db8::1234"
        # This should return True as the ipv6 address is within the specified range
        self.assertTrue(spf_check(spf_record, ipv6=ipv6_address))

        ipv6_address_outside = "2001:0db9::1234"
        # This should return False as the ipv6 address is outside the specified range
        self.assertFalse(spf_check(spf_record, ipv6=ipv6_address_outside))


if __name__ == "__main__":
    unittest.main()



import os
import random
import string
import unittest
from time import sleep
from typing import Optional

from dotenv import load_dotenv
from email_validator import EmailNotValidError
from mailinator import Mailinator, GetInboxRequest, GetMessageRequest

from smtpymailer.mailer import Contact, validate_send_email, SmtpMailer
from smtpymailer.utils import find_project_root


class TestContact(unittest.TestCase):
    def test_init_valid_email(self):
        """Test initializing a contact with a valid email."""
        contact = Contact("johndoe@example.com", "John Doe")
        self.assertEqual(contact.email, "johndoe@example.com")
        self.assertEqual(contact.name, "John Doe")


    def test_init_invalid_email(self):
        """Test initializing a contact with an invalid email."""
        with self.assertRaises(EmailNotValidError):
            contact = Contact("not a valid email", "John Doe")

    def test_repr_with_name(self):
        """Test the __repr__() method when name is present."""
        contact = Contact("johndoe@example.com", "John Doe", False)
        self.assertEqual(repr(contact), "John Doe <johndoe@example.com>")

    def test_repr_without_name(self):
        """Test the __repr__() method when name is not present."""
        contact = Contact("johndoe@example.com", None, False)
        self.assertEqual(repr(contact), "johndoe@example.com")

    def test_get_domain(self):
        """Test the get_domain() method."""
        contact = Contact("johndoe@example.com", "John Doe", False)
        self.assertEqual(contact.get_domain(), "example.com")


class TestValidateSendEmail(unittest.TestCase):
    recipient_domain = "team829298.testinator.com"
    email_subject = "My test email - "

    def setUp(self):
        root = find_project_root(file_path=__file__)
        load_dotenv(root.joinpath(".env"))

        # This method will run before each test
        self.root = find_project_root()
        self.assertIsNotNone(self.root, "Project root not found")

        self.template_path = os.path.join(self.root, "tests/templates")
        self.template_file = "test_real_email.html"
        self.full_template_path = os.path.join(self.template_path, self.template_file)
        self.html_content = "<html><body><h1>Test</h1><p>This is a test email with a unique id of $$ </p></body></html>"
        self.unique_id = "".join(random.choice(string.ascii_letters) for _ in range(10))
        self.unique_id_two = "".join(
            random.choice(string.ascii_letters) for _ in range(10)
        )
        self.unique_id_three = "".join(
            random.choice(string.ascii_letters) for _ in range(10)
        )
        self.attachments = ["./tests/assets/dog.pdf"]

    # Generate the random string

    def test_no_content_and_no_template(self):
        with self.assertRaises(ValueError):
            validate_send_email(None, None, None, "Subject", ["recipient@mail.com"])

    def test_no_subject(self):
        with self.assertRaises(ValueError):
            validate_send_email(
                "HTML content", None, None, None, ["recipient@mail.com"]
            )

    def test_no_recipients(self):
        with self.assertRaises(ValueError):
            validate_send_email("HTML content", None, None, "Subject", [])

    def test_invalid_recipients_email(self):
        with self.assertRaises(TypeError):
            validate_send_email(
                "HTML content",
                None,
                None,
                "Subject",
            )

    # Assuming is_file_with_path will return False
    def test_template_no_directory(self):
        with self.assertRaises(FileNotFoundError):
            validate_send_email(
                None, "template.txt", None, "Subject", ["recipient@mail.com"]
            )

    def test_valid(self):
        # Assuming all the inputs are valid, the test will pass if no Exceptions are raised
        validate_send_email(
            "HTML content", None, None, "Subject", ["recipient@mail.com"]
        )

    def test_template_with_list_of_paths(self):
        """Test validate_send_email with relative template file and list of template paths."""
        try:
            validate_send_email(
                None,
                self.template_file,
                [self.template_path, self.template_path],
                "Subject",
                ["recipient@mail.com"],
            )
        except Exception as e:
            self.fail(
                f"Test failed with relative template file and list of template paths: {e}"
            )

    def test_template_with_absolute_path(self):
        """Test validate_send_email with absolute template file path."""
        try:
            validate_send_email(
                None, self.full_template_path, None, "Subject", ["recipient@mail.com"]
            )
        except Exception as e:
            self.fail(f"Test failed with absolute template file path: {e}")

    def check_inbox_for_email(self, find_id: Optional[str] = None):
        """Checks the email inbox for a specific email.

        This method makes use of the Mailinator API to check the email inbox for a specific email. It uses the unique ID and recipient domain to filter the email messages. The method will make
        * multiple attempts with an increasing time wait interval to retrieve the email message.

        Parameters:
        - self: The instance of the class calling this method.
        - find_id: The unique ID to check for in the email subject. Optional. If not provided, the method will use the
            unique ID of the class instance.

        Returns:
        - If the email is found, the method returns the email message.
        - If the email is not found within the given number of attempts, the method returns None.

        Example usage:
        ```python
        email = check_inbox_for_email()
        if email is not None:
            print(email)
        ```
        """
        mailinator = Mailinator(os.getenv("MAILINATOR_API_KEY"))

        if find_id is None:
            find_id = self.unique_id

        time_wait = 2
        for i in range(0, 5):
            sleep(2)
            inbox = mailinator.request(
                GetInboxRequest(domain=self.recipient_domain, inbox=find_id, limit=5)
            )
            for message in inbox.msgs:
                # the id of the subject always remains the same
                if self.unique_id in message.subject:
                    return mailinator.request(
                        GetMessageRequest(
                            domain=self.recipient_domain,
                            inbox=find_id,
                            message_id=message.id,
                        )
                    )
            time_wait *= 2
        return None

    def send_test_email(self, **kwargs):
        """
        Sends a test email.

        Args:
            **kwargs: Additional keyword arguments to be passed to the `send_email` method.

        """
        mailer = SmtpMailer(os.getenv("MAIL_SENDER"), os.getenv("MAIL_SENDER_NAME"), **kwargs)
        return mailer.send_email(**kwargs)

    def check_for_id_in_message(self, message):
        """
        Checks if the given unique ID is found in the message text or HTML body.

        Args:
            self: The class instance.
            message: The message to check for the unique ID.

        Returns:
            A tuple with two boolean values indicating whether the unique ID was found in the text body
            and in the HTML body, respectively.
        """
        txt_body_found_id = False
        html_message_found_id = False
        for part in message.parts:
            if "text/plain" in part.headers.get("content-type"):
                if self.unique_id in part.body:
                    txt_body_found_id = True
            elif "text/html" in part.headers.get("content-type"):
                if self.unique_id in part.body:
                    html_message_found_id = True

        return txt_body_found_id, html_message_found_id

    def check_for_attachment_in_message(self, message, filepath, content_type):
        """
        Checks if the given unique ID is found in the message text or HTML body.

        Args:
            self: The class instance.
            message: The message to check for the unique ID.

        Returns:
            A tuple with two boolean values indicating whether the unique ID was found in the text body
            and in the HTML body, respectively.
        """
        file_name = os.path.basename(filepath)

        for part in message.parts:
            email_content_disposition = part.headers.get("content-disposition")
            email_content_type = part.headers.get("content-type")
            try:
                if (
                    file_name in email_content_disposition
                    and content_type == email_content_type
                ):
                    return True
            except TypeError:
                pass

    def test_send_email_with_html_text_to_one_recipient(self):
        """Test validate_send_email with html text."""

        email_kwargs = {
            "recipients": f"{self.unique_id}@{self.recipient_domain}",
            "subject": f"{self.email_subject}{self.unique_id}",
            "html_content": self.html_content.replace("$$", self.unique_id),
        }

        result = self.send_test_email(**email_kwargs)
        self.assertTrue(result, "Failed to send email")

        message = self.check_inbox_for_email()

        self.assertIsNotNone(message, "Email was not found in the inbox")

        txt_body_found_id, html_message_found_id = self.check_for_id_in_message(message)

        self.assertTrue(
            html_message_found_id, "Unique ID not found in the HTML body of the email"
        )
        self.assertTrue(
            txt_body_found_id, "Unique ID not found in the text body of the email"
        )

    def test_send_reply_to_in_email(self):
        """Test validate_send_email with html text."""

        email_kwargs = {
            "recipients": f"{self.unique_id}@{self.recipient_domain}",
            "reply_to": f"reply@{self.recipient_domain}",
            "subject": f"{self.email_subject}{self.unique_id}",
            "html_content": self.html_content.replace("$$", self.unique_id),
        }

        result = self.send_test_email(**email_kwargs)
        self.assertTrue(result, "Failed to send email")

        message = self.check_inbox_for_email()

        self.assertIsNotNone(message, "Email was not found in the inbox")

        txt_body_found_id, html_message_found_id = self.check_for_id_in_message(message)

        self.assertTrue(
            html_message_found_id, "Unique ID not found in the HTML body of the email"
        )
        self.assertTrue(
            txt_body_found_id, "Unique ID not found in the text body of the email"
        )

        self.assertTrue(
            f"reply@{self.recipient_domain}" == message.headers["reply-to"],
            "reply-to header not found in email",
        )

    def test_send_to_recipient_and_cc(self):
        """Test validate_send_email with html text."""

        email_kwargs = {
            "recipients": f"{self.unique_id}@{self.recipient_domain}",
            "cc_recipients": f"{self.unique_id_two}@{self.recipient_domain}",
            "subject": f"{self.email_subject}{self.unique_id}",
            "html_content": self.html_content.replace("$$", self.unique_id),
        }

        result = self.send_test_email(**email_kwargs)
        self.assertTrue(result, "Failed to send email")

        message = self.check_inbox_for_email()
        cc_message = self.check_inbox_for_email(self.unique_id_two)

        self.assertIsNotNone(message, "Email was not found in the inbox")
        self.assertIsNotNone(
            cc_message, "Email was not found in the Cc addresses inbox"
        )

        txt_body_found_id, html_message_found_id = self.check_for_id_in_message(message)

        self.assertTrue(
            html_message_found_id, "Unique ID not found in the HTML body of the email"
        )
        self.assertTrue(
            txt_body_found_id, "Unique ID not found in the text body of the email"
        )

    def test_send_to_recipient_with_template(self):
        """Test validate_send_email with template and jinja variables."""

        email_kwargs = {
            "recipients": f"{self.unique_id}@{self.recipient_domain}",
            "subject": f"{self.email_subject}{self.unique_id}",
            "template": self.full_template_path,
            "unique_id": self.unique_id,
        }

        result = self.send_test_email(**email_kwargs)
        self.assertTrue(result, "Failed to send email")

        message = self.check_inbox_for_email()

        self.assertIsNotNone(message, "Email was not found in the inbox")

        txt_body_found_id, html_message_found_id = self.check_for_id_in_message(message)

        self.assertTrue(
            html_message_found_id, "Unique ID not found in the HTML body of the email"
        )
        self.assertTrue(
            txt_body_found_id, "Unique ID not found in the text body of the email"
        )

    def test_send_to_recipient_with_attachment(self):
        """Test validate_send_email with template and jinja variables."""

        email_kwargs = {
            "recipients": f"{self.unique_id}@{self.recipient_domain}",
            "subject": f"{self.email_subject}{self.unique_id}",
            "template": self.full_template_path,
            "attachments": self.attachments,
            "unique_id": self.unique_id,
        }

        result = self.send_test_email(**email_kwargs)
        self.assertTrue(result, "Failed to send email")

        message = self.check_inbox_for_email()

        self.assertIsNotNone(message, "Email was not found in the inbox")

        txt_body_found_id, html_message_found_id = self.check_for_id_in_message(message)

        self.assertTrue(
            html_message_found_id, "Unique ID not found in the HTML body of the email"
        )
        self.assertTrue(
            txt_body_found_id, "Unique ID not found in the text body of the email"
        )

        self.check_for_attachment_in_message(
            message, self.attachments[0], "application/pdf"
        )

    def test_send_to_recipient_with_invalid_attachment(self):
        """Test validate_send_email with template and jinja variables."""

        email_kwargs = {
            "recipients": f"{self.unique_id}@{self.recipient_domain}",
            "subject": f"{self.email_subject}{self.unique_id}",
            "template": self.full_template_path,
            "attachments": "./assets/invalid.pdf",
            "unique_id": self.unique_id,
        }
        with self.assertRaises(FileNotFoundError):
            self.send_test_email(**email_kwargs)

    def test_failed_to_connect(self):
        email_kwargs = {
            "recipients": f"{self.unique_id}@{self.recipient_domain}",
            "subject": f"{self.email_subject}{self.unique_id}",
            "template": self.full_template_path,
            "attachments": self.attachments,
            "unique_id": self.unique_id,
            "MAIL_SERVER": "not.a.real.server",
        }

        with self.assertRaises(ValueError):
            self.send_test_email(**email_kwargs)


if __name__ == "__main__":
    unittest.main()



import os
from email import encoders
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.text import MIMEText
from mimetypes import guess_type
from pathlib import Path
from typing import Optional, Union, Tuple
from urllib.parse import urlparse

import validators

from smtpymailer.validation import validate_user_email

def is_get_local_file(file_path: str) -> Tuple[bool, Union[str, None]]:
    """
    Checks if the given file path is a local file.

    Parameters:
        file_path (str): The path of the file to check.

    Returns:
        Union[bool, Union[str, None]]: True and the absolute path of the file if it is a local file,
            False and None if it is not a local file.

    """

    abs_src = os.path.abspath(file_path)
    is_local = os.path.isfile(abs_src)
    if is_local:
        return True, abs_src
    else:
        return False, None


def is_resolvable(url):
    """
    Check if a URL is resolvable using validators library.

    Parameters:
        url (str): The URL to check if it is resolvable.

    Returns:
        bool: True if the URL is valid and resolvable, False otherwise.

    """
    try:
        # Use validators to check if the URL is valid
        return validators.url(url)
    except Exception:
        return False

def find_project_root(
    marker: Optional[Union[str, list]] = None, file_path: Optional = None
) -> Optional[Path]:
    """
    Traverse up from the current path until a directory containing the marker is found.

    Args:
        marker (Optional[Union[str, list]]): A string or a list of strings representing the file(s) or directory names
            to look for.If not provided, defaults to [".git", "setup.py", "requirements.txt", "README.md"].
        file_path (Optional[Path]): The __file__ of the calling module to start the search from. If not provided,
            defaults to the path of the current file.

    Returns:
        Optional[Path]: The path to the directory containing the marker, or None if not found.

    """
    if marker is None:
        marker = [".git", "setup.py", "requirements.txt", "README.md"]
    elif isinstance(marker, str):
        marker = [marker]

    current_file_path = (
        Path(file_path).resolve() if file_path else Path(__file__).resolve()
    )
    for parent in current_file_path.parents:
        if any((parent / m).exists() for m in marker):
            return parent

    return None


def is_file_with_path(path: str) -> bool:
    """
    Args:
        path: A string representing the path to the file.

    Return:
        A boolean value indicating whether the file exists at the specified path.

    Description:
        This method takes a path parameter and checks if the file exists at that path.
        It converts the given path to an absolute path using `os.path.abspath` function.
        Then, it uses `os.path.isfile` function to determine whether the absolute path refers to a file.
        The method returns `True` if the file exists at the specified path, and `False` otherwise.

    """
    absolute_path = os.path.abspath(path)
    return os.path.isfile(absolute_path)


def ensure_list(value: any) -> list:
    """
    Ensures that the given value is a list.

    Args:
        value (any): The recipient to be checked and converted to a list if necessary.

    Returns:
        list: The recipient as a list.

    """
    if value is None:
        return []
    return [value] if not isinstance(value, list) else value


def recipients_to_str(recipients):
    """
    Converts the recipients list to a string representation.

    Args:
        recipients: A list of recipients.

    Returns:
        A string representation of the recipients list, with each recipient separated by a comma and space.
        If the recipients list is empty, an empty string is returned.
    """
    return ", ".join(ensure_list(recipients)) if recipients else ""


def convert_bool(val):
    """
    Converts a string to a boolean value.

    Args:
        val (str): The string to convert.

    Returns:
        bool: True if the string is "True", False otherwise.
    """
    return val in ["True", "true", 1, "1", "Yes", "yes", "Y", "y", True]


def validate_and_extend(all_recipients: list, recipient_type: Union[str, list]):
    """
    Validates and extends the all_recipients list with validated email addresses. If a recipient type is provided
    as a string or list, it is first normalized into a list format. Each email address in this list is then validated.
    Valid email addresses are appended to the all_recipients list. If an invalid email is found, an error is raised.

    Args:
        all_recipients (list): The main list of all recipients that will be extended with validated emails.
        recipient_type (Union[str, list]): A string or list of email addresses to validate and add to all_recipients.

    Raises:
        EmailNotValidError: If any of the email addresses in recipient_type is not valid.

    """

    if recipient_type:
        recipient_list = ensure_list(recipient_type)
        for email in recipient_list:
            if validate_user_email(email):
                all_recipients.append(email)


def build_all_recipients_and_validate(
    recipients: Union[str, list],
    cc_recipients: Optional[Union[str, list]] = None,
    bcc_recipients: Optional[Union[str, list]] = None,
):
    """
    Constructs a complete list of validated email addresses for an email, including To, CC, and BCC recipients.
    It takes individual or lists of email addresses for each category, normalizes them, and then validates each email.
    All valid email addresses across these categories are compiled into a single list. An error is raised if any email
    address is found to be invalid.

    Args:
        recipients (Union[str, list]): A string or list of email addresses for the 'To' recipients.
        cc_recipients (Optional[Union[str, list]]): Optionally, a string or list of email addresses for
            the 'CC' recipients.
        bcc_recipients (Optional[Union[str, list]]): Optionally, a string or list of email addresses for
            the 'BCC' recipients.

    Returns:
        list: A list of all validated email addresses across the To, CC, and BCC fields.

    Raises:
        EmailNotValidError: If any of the provided email addresses are not valid.
    """
    if not len(ensure_list(recipients)) > 0:
        raise ValueError("No recipients provided")

    all_recipients = []
    for recipient_type in [recipients, cc_recipients, bcc_recipients]:
        validate_and_extend(all_recipients, recipient_type)

    return all_recipients

def guess_type_by_extension(filename):
    extension = filename.split('.')[-1].lower()

    # Extended mapping of file extensions to MIME types
    mime_mapping = {
        # Document types
        'docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
        'pdf': 'application/pdf',
        'txt': 'text/plain',
        'doc': 'application/msword',
        'ppt': 'application/vnd.ms-powerpoint',
        'pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
        'xls': 'application/vnd.ms-excel',
        'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        'odt': 'application/vnd.oasis.opendocument.text',
        'ods': 'application/vnd.oasis.opendocument.spreadsheet',
        'odp': 'application/vnd.oasis.opendocument.presentation',
        'rtf': 'application/rtf',
        'csv': 'text/csv',
        'html': 'text/html',
        'xml': 'application/xml',
        # Image types
        'jpg': 'image/jpeg',
        'jpeg': 'image/jpeg',
        'png': 'image/png',
        'gif': 'image/gif',
        'bmp': 'image/bmp',
        'svg': 'image/svg+xml',
        'tiff': 'image/tiff',
        'webp': 'image/webp',
        # Video types
        'mp4': 'video/mp4',
        'avi': 'video/x-msvideo',
        'mkv': 'video/x-matroska',
        'mov': 'video/quicktime',
        'wmv': 'video/x-ms-wmv',
        'flv': 'video/x-flv',
        'webm': 'video/webm',
        # Audio types
        'mp3': 'audio/mpeg',
        'wav': 'audio/wav',
        'ogg': 'audio/ogg',
        'flac': 'audio/flac',
        'aac': 'audio/aac',
        'm4a': 'audio/mp4',
        'wma': 'audio/x-ms-wma',
        # Archive types
        'zip': 'application/zip',
        'rar': 'application/x-rar-compressed',
        '7z': 'application/x-7z-compressed',
        'tar': 'application/x-tar',
        'gz': 'application/gzip',
        'bz2': 'application/x-bzip2',
        # Executable and script types
        'exe': 'application/x-msdownload',
        'sh': 'application/x-sh',
        'bat': 'application/x-bat',
        'py': 'text/x-python',
        'jar': 'application/java-archive',
        # Font types
        'woff': 'font/woff',
        'woff2': 'font/woff2',
        'ttf': 'font/ttf',
        'otf': 'font/otf',
        # Other
        'json': 'application/json',
        'md': 'text/markdown',
    }

    # Lookup MIME type based on extension
    mime_type = mime_mapping.get(extension)
    return mime_type

def construct_mime_object(path: str, attachment):
    """
    This function takes a file path and an attachment object, and constructs a MIME object
    that represents the attachment. This is particularly useful when dealing with different
    types of file attachments in email sending, as the constructed MIME object can be directly
    attached to the email object.

    Args:
        path (str): The path of the file to be attached. This path is used to guess the MIME type
            of the file, which helps in creating the appropriate type of MIME object.

        attachment (file object): The file object to be attached. The content of this file is
            read and placed in the payload of the MIME object.

    Returns:
        mime_part (object): The MIME object representing the attachment. This object can be
            differentiating depending on the MIME type of the input file, it could be of type.

                * MIMEApplication
                * MIMEText
                * MIMEImage
                * MIMEAudio
                * MIMEBase

            The payload of this object contains the content of the input file, and its 'Content-Disposition' header
            is set to designate it as an attachment with filename "example.txt".
    """

    mime_type = guess_type_by_extension(path)
    mime_main, mime_sub = mime_type.split("/")

    if mime_main == "application":
        mime_part = MIMEApplication(attachment.read(), _subtype=mime_sub)
    elif mime_main == "text":
        mime_part = MIMEText(
            attachment.read().decode("utf-8"), _subtype=mime_sub, _charset="utf-8"
        )
    elif mime_main == "image":
        mime_part = MIMEImage(attachment.read(), _subtype=mime_sub)
    elif mime_main == "audio":
        mime_part = MIMEAudio(attachment.read(), _subtype=mime_sub)
    else:
        mime_part = MIMEBase(mime_main, mime_sub)
        mime_part.set_payload(attachment.read())
        encoders.encode_base64(mime_part)

    mime_part.add_header("Content-Disposition", "attachment", filename="example.txt")
    return mime_part



import base64
import hashlib
import os
import unittest
from email.mime.multipart import MIMEMultipart
from unittest import mock
import smtpymailer.html_parse
from smtpymailer.utils import find_project_root
from smtpymailer.html_parse import (
    check_data_in_html_el,
    make_html_content,
    convert_img_elements
)


class TestRenderHtmlTemplate(unittest.TestCase):
    def test_render_full_path_html_template(self):
        """
        Tests the rendering of a given HTML template using the full path.

        Args:
            self (object): The instance of the test class.

        Returns:
            None

        Raises:
            AssertionError: If the rendered HTML does not match the expected HTML.

        """

        template_name = "test.html"
        full_path = os.path.join(find_project_root(), "tests/templates", template_name)

        expected_html = "<h1>foo</h1><h2>bar</h2>"

        result = smtpymailer.html_parse.render_html_template(
            template=full_path, foo="foo", bar="bar"
        )

        self.assertEqual(expected_html, result)

        result = smtpymailer.html_parse.render_html_template(
            template=full_path, foo="bar", bar="foor"
        )

        self.assertNotEqual(expected_html, result)

    def test_render_name_and_template(self):
        """
        Tests the rendering of a given HTML template using the full path.

        Args:
            self (object): The instance of the test class.

        Returns:
            None

        Raises:
            AssertionError: If the rendered HTML does not match the expected HTML.

        """

        template_name = "test.html"
        template_path = os.path.join(find_project_root(), "tests/templates")
        expected_html = "<h1>foo</h1><h2>bar</h2>"

        result = smtpymailer.html_parse.render_html_template(
            template=template_name, template_paths=template_path, foo="foo", bar="bar"
        )

        self.assertEqual(expected_html, result)

        result = smtpymailer.html_parse.render_html_template(
            template=template_name, template_paths=template_path, foo="bar", bar="foo"
        )

        self.assertNotEqual(expected_html, result)


class TestAttachRemoteImagesAsCid(unittest.TestCase):
    def attach_images_as_cid_helper(
        self, images, html_content=None, fake_content=None, fake_content_match=None
    ):
        if fake_content is None:
            fake_content = b"fake image content"
        if not fake_content_match:
            fake_content_match = fake_content

        with mock.patch("requests.get") as mock_get:
            # Set up the mock response
            mock_response = mock.Mock()
            mock_response.content = fake_content
            mock_response.raise_for_status = mock.Mock()
            mock_get.return_value = mock_response

            # Call the function with a test HTML string
            if html_content is None:
                image_elements = "".join(
                    ['<img data-cid src="https://example.com/image.jpg">'] * images
                )
                html_content = f"<html><body>{image_elements}</body></html>"

            msg = MIMEMultipart()
            result = convert_img_elements(html_content, msg)

            mock_get.assert_called_with("https://example.com/image.jpg", stream=True)

            cid_hash = hashlib.md5(fake_content_match).hexdigest()
            for i in range(images):
                self.assertIn(f"cid:{cid_hash}{i}", result)

            # Check the number of attachments
            self.assertEqual(len(msg.get_payload()), images)

            # Inspect the attachments
            image_numbers = [i for i in range(images)]
            imgno = 0
            for part in msg.walk():
                if part.get_content_maintype() == "image":
                    # Verify the Content-ID
                    self.assertIn(int(imgno), image_numbers)
                    # Verify the Content-Disposition
                    self.assertEqual(part.get("Content-Disposition"), "inline")
                    # Optionally, verify the image data
                    self.assertEqual(part.get_payload(decode=True), fake_content_match)
                    imgno += 1

            return result

    def test_attach_single_image_as_cid(self):
        self.attach_images_as_cid_helper(images=1)

    def test_attach_ten_images_as_cid(self):
        self.attach_images_as_cid_helper(images=10)

    def test_convert_cid_image_different_type(self):
        files = ["./tests/assets/1px.png", "../tests/assets/1px.png", "./smtpymailer/tests/assets/1px.png"]
        for file in files:
            if os.path.isfile(file):
                with open(file, "rb") as file:
                    png_content = file.read()
                break

        files = ["./tests/assets/1px.jpg", "../tests/assets/1px.jpg", "./smtpymailer/tests/assets/1px.jpg"]
        for file in files:
            if os.path.isfile(file):
                with open(file, "rb") as file:
                    jpg_content = file.read()


        html_content = '<html><body><img data-cid data-convert="png" data-format="RGB" src="https://example.com/image.jpg"></body></html>'
        self.attach_images_as_cid_helper(
            images=1,
            html_content=html_content,
            fake_content=jpg_content,
            fake_content_match=png_content,
        )

    def test_alter_img_html_with_none(self):
        msg = MIMEMultipart()
        html_content = "<html><body></body></html>"
        html_out = convert_img_elements(html_content, msg)
        self.assertEqual(html_content, html_out)

    @mock.patch("requests.get")
    def test_alter_img_html_with_base_and_cid(self, mock):
        # Set up the mock response
        mock_response = mock.Mock()
        mock_response.content = b"fake_img_data"
        mock_response.raise_for_status = mock.Mock()
        mock.return_value = mock_response

        msg = MIMEMultipart()
        html_content = (
            '<html><body><img data-cid src="https://example.com/image.jpg"></body></html>'
        )
        html_out = convert_img_elements(html_content, msg)
        self.assertTrue("cid:" in html_out)

        msg = MIMEMultipart()
        html_content = (
            '<html><body><img data-base src="https://example.com/image.jpg"></body></html>'
        )
        html_out = convert_img_elements(html_content, msg)
        self.assertTrue(";base64," in html_out)

    def test_make_html_content(self):
        html_content = "<h1>foo</h1><h2>bar</h2>"
        html_out = make_html_content(
            template="./tests/templates/test.html", foo="foo", bar="bar"
        )
        self.assertEqual(html_content, html_out)

        html_content = "<h1>foo</h1><h2>bar</h2>"
        html_out = make_html_content(template="test.html", template_directory="./tests/templates", foo="foo", bar="bar")
        self.assertEqual(html_content, html_out)

        html_content = "<h1>foo</h1><h2>bar</h2>"
        html_out = make_html_content(html_content=html_content)
        self.assertEqual(html_content, html_out)

        html_content = "<h1>foo</h1><h2>bar</h2>"
        html_out = make_html_content(
            template="./tests/templates/test.html", foo="foo", bar="baz"
        )
        self.assertNotEqual(html_content, html_out)

    def test_convert_cid_image_same_type(self):
        html_content = '<html><body><img data-cid data-convert="jpg" data-format="RGB" src="https://example.com/image.jpg"></body></html>'
        self.attach_images_as_cid_helper(images=1, html_content=html_content)

    def test_fail_convert_cid_type(self):
        with self.assertRaises(ValueError):
            html_content = '<html><body><img data-cid  data-convert="webp" data-format="RGB" src="https://example.com/image.jpg"></body></html>'
            self.attach_images_as_cid_helper(images=1, html_content=html_content)

    def test_fail_convert_cid_format(self):
        with self.assertRaises(ValueError):
            html_content = '<html><body><img data-cid  data-convert="jpg" data-format="cmyk" src="https://example.com/image.jpg"></body></html>'
            self.attach_images_as_cid_helper(images=1, html_content=html_content)

    def test_attach_unavailable_url_cid(self):
        html_content = (
            '<html><body><img data-cid src="htts://example.com/image.jpg"></body></html>'
        )
        msg = MIMEMultipart()
        result = convert_img_elements(html_content, msg)

        self.assertNotIn(f"cid:image0.jpg", result)

        # Check the number of attachments
        self.assertEqual(len(msg.get_payload()), 0)


class TestConvertRemoteImgElementsToBase64(unittest.TestCase):
    def convert_images_to_base64_helper(self, images):
        """
        Converts images in HTML content to base64 format.

        Args:
            images (int): The number of images to convert to base64.

        """
        with mock.patch("requests.get") as mock_get:
            # Set up the mock response
            mock_response = mock.Mock()
            mock_response.content = b"fake image content"
            mock_response.raise_for_status = mock.Mock()
            mock_get.return_value = mock_response

            # Call the function with a test HTML string
            image_elements = "".join(
                ['<img data-base src="https://example.com/image.jpg">'] * images
            )
            html_content = f"<html><body>{image_elements}</body></html>"
            result = convert_img_elements(html_content, MIMEMultipart())

            converted_data = check_data_in_html_el(result)

            self.assertEqual(converted_data["base64"], images)
            self.assertEqual(converted_data["cid"], 0)
            self.assertEqual(converted_data["data-smtpymailer"], images)
            self.assertEqual(mock_get.call_count, images)

    def test_convert_single_image_to_base64(self):
        self.convert_images_to_base64_helper(images=1)

    def test_convert_ten_images_to_base64(self):
        self.convert_images_to_base64_helper(images=10)

    def test_attach_unavailable_url_base64(self):
        html_content = (
            '<html><body><img data-base src="https://example.com/image.jpg"></body></html>'
        )
        result = convert_img_elements(html_content, MIMEMultipart())

        converted_data = check_data_in_html_el(result)

        self.assertEqual(converted_data["base64"], 0)
        self.assertEqual(converted_data["cid"], 0)
        self.assertEqual(converted_data["data-smtpymailer"], 0)

    def test_attach_invalid_url_base64(self):
        html_content = '<html><body><img data-base src="example.com/image.jpg"></body></html>'
        result = convert_img_elements(html_content, MIMEMultipart())

        converted_data = check_data_in_html_el(result)

        self.assertEqual(converted_data["base64"], 0)
        self.assertEqual(converted_data["cid"], 0)
        self.assertEqual(converted_data["data-smtpymailer"], 0)

class TestConvertLocalImgElementsToBase64(unittest.TestCase):

    def read_image_base64(self, image_path):
        """Reads an image and converts it to base64."""
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def test_convert_single_image_to_base64(self):
        html_content = '<html><body><img data-base src="./tests/assets/dog.jpg"></body></html>'
        result = convert_img_elements(html_content, MIMEMultipart())

        # Assuming the function replaces the src attribute with the base64 data
        expected_base64_data = self.read_image_base64('./tests/assets/dog.jpg')
        self.assertIn(expected_base64_data, result)

    def test_attach_invalid_url_base64(self):
        # This test checks if the function properly handles invalid image paths
        html_content = '<html><body><img data-base src="invalid_path/dog.jpg"></body></html>'
        result = convert_img_elements(html_content, MIMEMultipart())

        # Assuming the function leaves the src attribute unchanged if the image can't be read
        self.assertIn('src="invalid_path/dog.jpg"', result)

class TestAttachLocalImagesAsCid(unittest.TestCase):

    def attach_images_as_cid_helper(
        self, images, html_content=None, fake_content=None, fake_content_match=None
    ):
        if not fake_content_match:
            # Read the local image file
            with open('./tests/assets/dog.jpg', 'rb') as file:
                fake_content_match = file.read()

        if html_content is None:
            image_elements = "".join(
                ['<img data-cid src="./tests/assets/dog.jpg">'] * images
            )
            html_content = f"<html><body>{image_elements}</body></html>"

        msg = MIMEMultipart()
        result = convert_img_elements(html_content, msg)

        cid_hash = hashlib.md5(fake_content_match).hexdigest()
        for i in range(images):
            self.assertIn(f"cid:{cid_hash}{i}", result)

        # Check the number of attachments
        self.assertEqual(len(msg.get_payload()), images)

        # Inspect the attachments
        image_numbers = [i for i in range(images)]
        imgno = 0
        for part in msg.walk():
            if part.get_content_maintype() == "image":
                # Verify the Content-ID
                self.assertIn(int(imgno), image_numbers)
                # Verify the Content-Disposition
                self.assertEqual(part.get("Content-Disposition"), "inline")
                # Optionally, verify the image data
                self.assertEqual(part.get_payload(decode=True), fake_content_match)
                imgno += 1

        return result

    def test_attach_single_image_as_cid(self):
        self.attach_images_as_cid_helper(images=1)

    def test_attach_lots_of_images_as_cid(self):
        self.attach_images_as_cid_helper(images=10)

if __name__ == "__main__":
    unittest.main()



from setuptools import setup

# Reading long description from README
with open("README.md", "r", encoding='UTF-8') as fh:
    long_description = fh.read()

# Reading requirements from requirements.txt
with open("requirements.txt", "r", encoding='UTF-8') as fh:
    requirements = fh.read().splitlines()
with open("./tests/requirements-test.txt", "r", encoding='UTF-8') as fh:
    test_requirements = fh.read().splitlines()

setup(
    name="smtpymailer",
    version="0.1.1",
    author="Lewis Morris",
    author_email="lewis@arched.dev",
    description="A emailing python library for emailing from alternative domains - DNS entries etc need to be assigned (not used for spamming).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/lewis-morris/smtpymailer",
    packages=["smtpymailer"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=requirements,
    extras_require={
        'dev': test_requirements
    },
    include_package_data=True,
)



import base64
import binascii
import ipaddress
import re
import socket
from typing import Optional, Union, Tuple, List

import dns.resolver
import validators
from email_validator import validate_email


def validate_user_email(
    email: str, return_normalized: bool = True, check_deliverability: bool = False
):
    """
    Validates a user email address.

    Args:
        email (str): The email address to be validated.
        return_normalized (bool): Specifies whether to return the normalized email or not.
                                  Default is True. If set to True, the normalized email will be returned.
        check_deliverability (bool): Specifies whether to check the deliverability of the email address or not.
                                     Default is False. If set to True, the deliverability of the email address will be checked.

    Returns:
        str: The normalized email address, if return_normalized is set to True.
        OR
        str: The original email address.

    """
    email_info = validate_email(email, check_deliverability=check_deliverability)
    return email_info.normalized if return_normalized else email


def validate_dkim_record(dkim_record: str) -> Union[Optional[bytes], bool]:
    """
    Args:
        dkim_record: A string representing the DKIM record that needs to be validated.

    Returns:
        Union[Optional[bytes], bool]: Returns either the decoded public key as bytes or False.

    """

    dkim_regex = re.compile(
        r"v=DKIM1;(\s*h=sha(1|256);)?(\s*k=rsa;)?(\s*t=[\w/]+;)?(\s*p=[A-Za-z0-9+/]+={0,2})"
    )

    # Remove quotes and spaces from the record for validation
    dkim_record = dkim_record.replace('"', "").replace(" ", "").strip()

    # Matching the record with the regex
    match = dkim_regex.fullmatch(dkim_record)
    if not match:
        return False

    # Extract and decode the public key
    public_key_encoded = match.group(5)[2:]  # Correctly remove 'p='

    return True if base64.b64decode(public_key_encoded) else False


def get_address_type(address):
    """
    Takes a string and returns a dictionary indicating whether it's an IPv4, IPv6, or domain.

    Args:
    address (str): The address to be checked.

    Returns:
    dict: A dictionary with the address type and the address.
    """
    try:
        # Check for IPv4
        ipaddress.IPv4Address(address)
        return {"ipv4": address}
    except ipaddress.AddressValueError:
        pass

    try:
        # Check for IPv6
        ipaddress.IPv6Address(address)
        return {"ipv6": address}
    except ipaddress.AddressValueError:
        pass

    # If it's not an IP address, treat it as a domain
    if validators.domain(address):
        return {"domain": address}

    return {"invalid": address}


def resolve_domain(domain: str) -> Tuple[List[str], List[str]]:
    """
    Resolve a domain to its corresponding IP addresses.

    Args:
        domain (str): The domain name to resolve.

    Returns:
        tuple: A tuple containing two lists. The first list contains IPv4 addresses
               associated with the domain, while the second list contains IPv6 addresses.
    """
    ipv4_addresses = []
    ipv6_addresses = []

    for info in socket.getaddrinfo(domain, None):
        address = info[4][0]
        if info[0] == socket.AF_INET:
            # IPv4 address
            ipv4_addresses.append(address)
        elif info[0] == socket.AF_INET6:
            # IPv6 address
            ipv6_addresses.append(address)

    return ipv4_addresses, ipv6_addresses


def is_ip_in_network(ip: str, network: str) -> bool:
    """
    Check if an IP address is in a given network.

    Args:
        ip (str): The IP address to check.
        network (str): The network to check against.

    Returns:
        bool: True if the IP address is in the network, False otherwise.

    Raises:
        TypeError: If the IP address or network is invalid.
        ValueError: if the IP or network is not a string.

    """
    if not isinstance(ip, str) or not isinstance(network, str):
        raise TypeError("Invalid input parameters")

    try:

        ip_obj = ipaddress.ip_address(ip)
        network_obj = ipaddress.ip_interface(network).network
        return ip_obj in network_obj

    except ValueError:
        raise ValueError("Invalid IP or network")


def spf_check(
    spf_record: str,
    ipv4: Optional[Union[str, List[str]]] = None,
    ipv6: Optional[Union[str, List[str]]] = None,
    domain: Optional[str] = None,
):
    """
    Args:
        spf_record: The SPF record to be checked.
        ipv4: Optional parameter for the IPv4 addresses to be checked against the SPF record. It can be a single
            IPv4 address or a list of IPv4 addresses.
        ipv6: Optional parameter for the IPv6 addresses to be checked against the SPF record. It can be a single
            IPv6 address or a list of IPv6 addresses.
        domain: Optional parameter for the domain name to be checked against the SPF record.

    Returns:
        Returns True if the SPF record authorizes the provided IPv4 addresses, IPv6 addresses, or the domain.
        Returns False if the SPF record explicitly disallows them.

    Raises:
        ValueError: If the SPF record is invalid.

    """

    def list_fix(
        val: Union[List, str], add: Optional[Union[List, str]] = None
    ) -> Optional[List]:
        """
        Args:
            val: The input value that needs to be fixed. It can be a single value or a list of values.
            add: An optional value to be added to the input value. It can be a single value or a list of values.

        Returns:
            A list with unique elements from the input value and the additional value. If the resulting list is empty, it returns None.

        Example Usage:
            val = [1, 2, 3]
            add = [2, 3, 4, 5]
            fixed_list = list_fix(val, add)
        """
        val = [val] if not isinstance(val, list) else val
        add = [add] if add and not isinstance(add, list) else add
        val += add or []

        return_val = list(set(val))
        return return_val if return_val else []

    if (not ipv4 or not ipv6) and domain:
        _ipv4, _ipv6 = resolve_domain(domain)
        ipv4 = list_fix(_ipv4)
        ipv6 = list_fix(_ipv6)

    # Remove quotes and strip whitespace
    spf_record = spf_record.strip('"').strip()

    # Split the SPF record into parts
    parts = spf_record.split()

    # Check if the SPF record is valid
    if not parts or parts[0] != "v=spf1":
        raise ValueError("Invalid SPF record")

    # Check each part of the SPF record
    for part in parts[1:]:
        if part.startswith("include:"):
            included_domain = part.split(":")[1]
            if domain == included_domain:
                return True
            try:
                answers = dns.resolver.resolve(included_domain, "TXT")
                for rdata in answers:
                    if spf_check(str(rdata), ipv4, ipv6, domain):
                        return True
            except Exception as e:
                return False

        elif part.startswith("ip4:"):
            network = part.split(":")[1]
            # Convert ipv4 to a list if it's not already
            ipv4_list = [ipv4] if isinstance(ipv4, str) else ipv4
            for ipv4_address in ipv4_list:
                if ipv4_address and is_ip_in_network(ipv4_address, network):
                    return True

        elif part.startswith("ip6:"):
            network = ":".join(part.split(":")[1:])
            # Convert ipv6 to a list if it's not already
            ipv6_list = [ipv6] if isinstance(ipv6, str) else ipv6
            for ipv6_address in ipv6_list:
                if ipv6_address and is_ip_in_network(ipv6_address, network):
                    return True

    # Check the -all mechanism
    if "-all" in parts:
        return False  # The IP or domain is not authorized

def get_dmarc_record_match(record):
    """
    Clean up the record and checks if it's a match.

    Args:
        record: A string containing the DMARC record.

    Returns:
        A Match object if the record is a valid DMARC record.
        Throws ValueError if regex fullmatch fails.

    Raises:
        EmailNotValidError: If any of the email addresses in mailto: are not valid.
    """
    # Removing quotes and spaces from the record for validation
    clean_record = record.replace('"', "").strip()

    mailto_addresses = re.findall(r"mailto:([^;]+)", clean_record)
    for address in mailto_addresses:
        validate_user_email(address)

    # Regular expression to validate DMARC record
    dmarc_regex = re.compile(
        r"^v=DMARC1;\s*((p=none|p=quarantine|p=reject|rua=mailto:[^;]+|ruf=mailto:[^;]+|pct=\d{1,3}|sp=none|sp=quarantine|sp=reject|aspf=r|aspf=s|adkim=r|adkim=s|fo=[01ds]|rf=afrf|rf=iodef|ri=\d+);?\s*)*\s*$",
        re.IGNORECASE,
    )
    return dmarc_regex.fullmatch(clean_record)


def validate_dmarc_record(record):
    """
    Validates a DMARC record.

    Args:
        record: A string containing the DMARC record.

    Returns:
        A Match object if the record is a valid DMARC record.
        Throws ValueError if the record is not a valid DMARC record.
    """
    # Check if record is a string
    assert isinstance(record, str), "The record must be a string."

    # Clean up and match the record
    record_match = get_dmarc_record_match(record)

    # Check pct value if exists
    if record_match:
        pct_match = re.search(r"pct=(\d{1,3})", record)
        if pct_match:
            pct_value = int(pct_match.group(1))
            if pct_value < 0 or pct_value > 100:
                raise ValueError("The pct value must be between 0 and 100.")
    else:
        raise ValueError("The record is not a valid DMARC record.")

    return True if record_match else False
